import { WebSocket } from 'ws';
import { v4 as uuidv4 } from 'uuid';
import { AnyObj, WSAnyEndpoints, WSPacket, WSPacketError } from 'wetube-common';

export interface WSClientConnectionOptions {
    onOpen?: (connection: WSClientConnection) => void,
    onClose?: (connection: WSClientConnection, code: number, reason: Buffer) => void,
    onMessage: (message: WSPacket, connection: WSClientConnection) => Promise<AnyObj | null>
}
export class WSClientConnection {
    private _id: string;
    private ws: WebSocket;
    private options: WSClientConnectionOptions;

    public authToken: string | null = null;

    constructor(ws: WebSocket, options: WSClientConnectionOptions) {
        this._id = uuidv4();
        this.options = options;
        this.ws = ws;

        this.ws.once('open', this.onOpen);
        this.ws.once('close', this.onClose);
        this.ws.on('message', this.onMessage);
    }

    public get id(): string {
        return this._id;
    }

    public send<T extends WSAnyEndpoints, K extends keyof T>(type: K, data?: T[K]['response'] | WSPacketError, ackId?: number) {
        let dataObj = data ?? {};
        let errorObj = undefined;
        if (this.isErrorData(data)) {
            dataObj = {};
            errorObj = data;
        }

        const message = JSON.stringify({
            type,
            body: {
                id: this._id,
                data: dataObj,
                error: errorObj,
                ackId,
            }
        });

        this.ws.send(message);
    }

    private onOpen = () => {
        if (this.options.onOpen) {
            this.options.onOpen(this);
        }
    };

    private onClose = (code: number, reason: Buffer) => {
        if (this.options.onClose) {
            this.options.onClose(this, code, reason);
        }

        this.ws.off('message', this.onMessage);
    };

    private onMessage = async (message: string) => {
        const parsedMessage: WSPacket = JSON.parse(message);
        if (parsedMessage.type == '.handshake') {
            this.onHandshake(parsedMessage.body.ackId);
            return;
        }

        if (this.options.onMessage) {
            const data = await this.options.onMessage(parsedMessage, this);

            // data == null means we don't want to send ack
            if (!data || !parsedMessage.body.ackId) {
                return;
            }

            this.send('.ack', data ?? undefined, parsedMessage.body.ackId);
        }
    };

    private onHandshake = (ackId?: number) => {
        this.send('.ack', {
            id: this._id,
        }, ackId);
    };

    private isErrorData = (data: unknown): data is WSPacketError => {
        return typeof data == 'object' && data != null &&
            'code' in data && typeof data.code == 'number';
    };

}