import { WebSocket, Server as WebSocketServer } from 'ws';
import { WSPacket, WSPacketError } from 'wetube-common';
import { WSClientConnection } from './WSClientConnection';
import { WSController, WSAnyEndpoints, WSEndpointBindObj, WSEndpointRequest } from './WSController';

export class WSRouter {
    private connections: WSClientConnection[] = [];
    private controllers: WSController[] = [];
    private handlers = new Map<string, WSEndpointBindObj>();
    private server: WebSocketServer;

    constructor(server: WebSocketServer) {
        this.server = server;
    }

    public register<T extends WSAnyEndpoints>(controller: WSController<T>) {
        const handlers = controller.register();
        for (const message in handlers) {
            const handler = handlers[message];
            if (!handler) {
                throw new Error(`[WebSocketRouter] Missing handler for message "${message}"`);
            }
            if (this.handlers.has(message)) {
                throw new Error(`[WebSocketRouter] Message "${message}" is already handled by another controller`);
            }

            // intentional
            this.handlers.set(message, handler as unknown as WSEndpointBindObj);
        }

        // intentional
        this.controllers.push(controller as unknown as WSController);
    }

    public start() {
        this.server.on('connection', (ws: WebSocket) => {
            const clientConnection = new WSClientConnection(ws, {
                onClose: (connection) => {
                    this.onDisconnect(connection);
                },
                onMessage: (message, connection) => {
                    return this.handle(message, connection);
                }
            });
            this.connections.push(clientConnection);
            this.onConnection(clientConnection);
        });

    }

    private async handle(msg: WSPacket, connection: WSClientConnection) {
        console.log(`Received message: ${JSON.stringify(msg)}`);

        const handler = this.handlers.get(msg.type);
        if (!handler) {
            console.log(`unexpected message type: ${String(msg.type)}`);
            return null;
        }

        const middleware = handler.middleware ? [...handler.middleware] : [];
        const request: WSEndpointRequest = { body: msg.body };
        const executeMiddlewareChain = async (error?: WSPacketError) => {
            if (error) {
                return error;
            }

            const nextMiddleware = middleware.shift();
            if (nextMiddleware) {
                await nextMiddleware(request, connection, executeMiddlewareChain);
            }
        };
        const error = await executeMiddlewareChain();
        if (error) {
            return error;
        }

        return handler.handler(request, connection);
    }

    private onConnection = (connection: WSClientConnection) => {
        this.controllers.forEach(c => c.onConnection(connection));
    };

    private onDisconnect = (connection: WSClientConnection) => {
        this.controllers.forEach(c => c.onDisconnect(connection));
    };
}