import { LobbyAPI, LobbyMessageType } from 'wetube-common';
import { ILobbyService } from '../services/lobby';
import { IAuthDataServices } from '../services/auth-data';
import {
    IMediaSyncService, SyncPlaylistPayload,
    PlayMediaPayload, PauseMediaPayload
} from '../services/media-sync';
import { IMediaMetadataService, MediaMetadata } from '../services/media-metadata';
import { ISubscription, Subscription, WSRes } from '../utils';

export type AppApiSubscriptions = {
    'join-lobby': WSRes<LobbyAPI, LobbyMessageType.JOIN>,
    'leave-lobby': WSRes<LobbyAPI, LobbyMessageType.LEAVE>,
    'sync-playlist-media': SyncPlaylistPayload,
    'play-media': PlayMediaPayload,
    'pause-media': PauseMediaPayload,
}

export interface AppApiServices {
    lobbyService: ILobbyService,
    authDataService: IAuthDataServices,
    mediaSyncService: IMediaSyncService,
    mediaMetadataService: IMediaMetadataService,
}
export class AppApi {
    private _subscription: Subscription<AppApiSubscriptions>;
    private _userId: string | undefined;
    private services: AppApiServices;

    constructor(services: AppApiServices) {
        this._subscription = new Subscription();
        this.services = services;

        this.services.lobbyService.subscription
            .subscribe('join', (data) => this._subscription.publish('join-lobby', data))
            .subscribe('leave', (data) => this._subscription.publish('leave-lobby', data));

        this.services.mediaSyncService.subscription
            .subscribe('sync-playlist', (data) => this._subscription.publish('sync-playlist-media', data))
            .subscribe('play', (data) => this._subscription.publish('play-media', data))
            .subscribe('pause', (data) => this._subscription.publish('pause-media', data));
    }

    public get subscription(): ISubscription<AppApiSubscriptions> {
        return this._subscription;
    }

    public get userId(): string | undefined {
        return this._userId;
    }

    public createLobby() {
        return this.services.lobbyService.create();
    }

    public joinLobby(lobbyId: string) {
        return this.services.lobbyService.join(lobbyId)
            .then(res => {
                const { data } = res;
                if (!data || !data.user) {
                    return Promise.reject();
                }

                this._userId = data.user.id;
                this.services.mediaSyncService.initialize(data.lobbyId, data.users, data.user.id);

                return res;
            });
    }

    public leaveLobby(lobbyId: string) {
        this.services.lobbyService.leave(lobbyId);
    }

    public getMediaMetadata(query: string) {
        return this.services.mediaMetadataService.getMetadataByQuery(query);
    }

    public syncMediaPlaylist(playlist: MediaMetadata[], targetUserId?: string) {
        this.services.mediaSyncService.syncPlaylist(playlist, targetUserId);
    }

    public playMedia(timestamp: number) {
        this.services.mediaSyncService.play(timestamp);
    }

    public pauseMedia(timestamp: number) {
        this.services.mediaSyncService.pause(timestamp);
    }

}