import React, { useState } from 'react';
import { styled } from '@mui/material/styles';
import Box from '@mui/material/Box';
import IconButton from '@mui/material/IconButton';
import Icon from '@mui/material/Icon';
import Card from '@mui/material/Card';
import CardContent from '@mui/material/CardContent';
import CardMedia from '@mui/material/CardMedia';
import Typography from '@mui/material/Typography';
import { MediaMetadata } from '../../../services/media-metadata';
import Swipeout from 'rc-swipeout';
import 'rc-swipeout/assets/index.css';

const StyledCard = styled(Card)({
    display: 'flex',
    width: '100%'
});

const StyledCardMedia = styled(CardMedia)({
    width: 150,
    height: 84,
});

const StyledCardContent = styled(CardContent)({
    minWidth: 218,
    flex: 1,
});

interface VideoItemProps {
    className?: string,
    metadata: MediaMetadata,
}
export const VideoItem = (props: VideoItemProps) => {
    const { metadata, className } = props;

    return (
        <StyledCard className={className}>
            <StyledCardMedia
                component="img"
                image={metadata.thumbnailUrl}
            />
            <StyledCardContent>
                <Typography
                    noWrap
                    variant="body1"
                    component="div"
                >
                    {metadata.title}
                </Typography>
                <Typography
                    noWrap
                    variant="body2"
                    color="text.secondary"
                    component="div"
                >
                    {metadata.author}
                </Typography>
            </StyledCardContent>
        </StyledCard>
    );
};

const ClickableBox = styled(Box)({
    cursor: 'pointer',
});

interface ClickableVideoItemProps {
    className?: string,
    metadata: MediaMetadata | null,
    onClick: (metadata: MediaMetadata) => void;
}
export const ClickableVideoItem = ({ className, metadata, onClick }: ClickableVideoItemProps) => {
    return metadata ? (
        <ClickableBox
            className={className}
            onClick={() => onClick(metadata)}
        >
            <VideoItem
                metadata={metadata}
            />
        </ClickableBox>
    ) : null;
};

const Overlay = styled('div')({
    userSelect: 'none',
    position: 'absolute',
    backgroundColor: 'rgba(0, 0, 0, 0.5)',
    top: 0,
    bottom: 0,
    left: 0,
    right: 0,
    marginTop: 8,
    marginBottom: 8,
    marginLeft: 16,
    marginRight: 16,
    display: 'flex',
    justifyContent: 'flex-start',
    alignItems: 'center',
});
const StyledIconButton = styled(IconButton)({
    margin: 25,
});

const Swipe = Swipeout as any;
interface DeletableVideoItemProps {
    className?: string,
    metadata: MediaMetadata | null,
    onClick: (metadata: MediaMetadata) => void;
    onDelete: (metadata: MediaMetadata) => void;
}
export const DeletableVideoItem = ({ className, metadata, onClick, onDelete }: DeletableVideoItemProps) => {
    const [showOverlay, setShowOverlay] = useState(false);

    return metadata ? (
        <div
            className={className}
            onMouseEnter={() => setShowOverlay(true)}
            onMouseLeave={() => setShowOverlay(false)}
            onClick={() => onClick(metadata)}
        >
            <Swipe
                right={[{
                    text: (<Icon>delete</Icon>),
                    onPress: () => onDelete(metadata),
                    style: { color: 'white', borderRadius: '25px' },
                }]}
            >
                <VideoItem
                    metadata={metadata}
                />
            </Swipe>
            {showOverlay ?
                <Overlay>
                    <StyledIconButton onClick={() => onClick(metadata)}>
                        <Icon>play_arrow</Icon>
                    </StyledIconButton>
                    <IconButton onClick={() => onDelete(metadata)}>
                        <Icon>delete</Icon>
                    </IconButton>
                </Overlay>
                : null}
        </div>
    ) : null;
};
