import { useState, useCallback, useRef, useEffect } from 'react';
import YouTube, { YouTubeEvent } from 'react-youtube';
import { AppApiSubscriptions } from '../../../api';
import { useAppApi } from '../../../context';

export const usePlayer = () => {
    const api = useAppApi();
    const playerRef = useRef<YouTube>(null);
    const [playing, setPlaying] = useState(false);
    const [duration, setDuration] = useState(0);
    const [elapsed, setElapsed] = useState(0);
    const [volume, setVolume] = useState(0);

    // Initial setup

    useEffect(() => {
        const playMedia = (data: AppApiSubscriptions['play-media']) => {
            playerRef.current?.internalPlayer.seekTo(data.currentTime, true);
            playerRef.current?.internalPlayer.playVideo();
        };

        const pauseMedia = (data: AppApiSubscriptions['pause-media']) => {
            playerRef.current?.internalPlayer.pauseVideo();
            playerRef.current?.internalPlayer.seekTo(data.currentTime, true);
        };

        api.subscription
            .subscribe('play-media', playMedia)
            .subscribe('pause-media', pauseMedia);

        const timer = setInterval(() => {
            if (playerRef.current) {
                playerRef.current?.internalPlayer.getCurrentTime().then((result: number) => {
                    if (result != undefined) {
                        setElapsed(result);
                    }
                });
            }
        }, 200);

        return () => {
            clearInterval(timer);
            api.subscription
                .unsubscribe('play-media', playMedia)
                .unsubscribe('pause-media', pauseMedia);
        };
    }, []);

    // Player callbacks

    const onPlayerPlay = useCallback(() => {
        setPlaying(true);
    }, []);

    const onPlayerPause = useCallback(() => {
        setPlaying(false);
    }, []);

    const onPlayerEnd = useCallback(() => {
        setPlaying(false);
    }, []);

    const onPlayerStateChange = useCallback((event: YouTubeEvent<number>) => {
        const duration = event.target.getDuration();
        setDuration(duration);

        const volume = event.target.getVolume();
        setVolume(volume);
    }, []);

    // Controls callbacks

    const onControlsPlay = useCallback(() => {
        const sec = Math.round(elapsed);

        api.playMedia(sec);
        playerRef.current?.internalPlayer.seekTo(sec, true);
        playerRef.current?.internalPlayer.playVideo();
    }, [elapsed]);

    const onControlsPause = useCallback(() => {
        const sec = Math.round(elapsed);

        api.pauseMedia(sec);
        playerRef.current?.internalPlayer.pauseVideo();
        playerRef.current?.internalPlayer.seekTo(sec, true);
    }, [elapsed]);

    const onControlsSeek = useCallback((value: number | number[]) => {
        if (Array.isArray(value)) {
            return;
        }
        const sec = Math.round(duration * (value / 100));
        api.playMedia(sec);

        playerRef.current?.internalPlayer.seekTo(sec, true);
        playerRef.current?.internalPlayer.playVideo();
    }, [duration]);

    const onControlsVolumeChange = useCallback((value: number | number[]) => {
        if (Array.isArray(value)) {
            return;
        }

        playerRef.current?.internalPlayer.setVolume(value).then(() => {
            playerRef.current?.internalPlayer.getVolume().then((val: number) => {
                setVolume(val);
            });
        });
    }, []);

    return {
        playing,
        duration,
        elapsed,
        volume,
        playerRef,
        onPlayerPlay,
        onPlayerPause,
        onPlayerEnd,
        onPlayerStateChange,
        onControlsPlay,
        onControlsPause,
        onControlsSeek,
        onControlsVolumeChange,
    };
};