import { LobbyUserObj } from 'wetube-common';
import { IMediaSyncService, MediaSyncSubscriptions } from './IMediaSyncService';
import { MediaMetadata } from '../media-metadata';
import { ILobbyService, LobbySubscriptions } from "../lobby";
import { ISubscription, Subscription } from '../../utils';

interface PeerConnectionClient {
    id: string,
    connection: RTCPeerConnection,
    channel: RTCDataChannel,
    messageQueue: string[],
}

export class MediaSyncService implements IMediaSyncService {
    private _subscription: Subscription<MediaSyncSubscriptions>;
    private clients = new Map<string, PeerConnectionClient>();
    private userId: string | null = null;
    private lobbyService: ILobbyService;

    constructor(lobbyService: ILobbyService) {
        this._subscription = new Subscription();
        this.lobbyService = lobbyService;
    }

    public get subscription(): ISubscription<MediaSyncSubscriptions> {
        return this._subscription;
    }

    public initialize(lobbyId: string, users: LobbyUserObj[], currentUserId: string) {
        this.lobbyService.subscription
            .subscribe('join', this.onLobbyJoin)
            .subscribe('leave', this.onLobbyLeave)
            .subscribe('data-offer', this.onDataOffer)
            .subscribe('data-answer', this.onDataAnswer)
            .subscribe('ice-candidate', this.onIceCandidate);

        this.createClients(lobbyId, users, currentUserId);
    }

    public syncPlaylist(playlist: MediaMetadata[], targetUserId?: string) {
        if (targetUserId && this.clients.has(targetUserId)) {
            const client = this.clients.get(targetUserId)!;
            this.send(client, 'sync-playlist', {
                playlist
            });
        } else {
            this.broadcast('sync-playlist', {
                playlist,
            });
        }
    }

    public play(currentTime: number) {
        this.broadcast('play', {
            currentTime,
        });
    }

    public pause(currentTime: number) {
        this.broadcast('pause', {
            currentTime,
        });
    }

    private broadcast(type: string, data?: object) {
        this.clients.forEach(c => {
            this.send(c, type, data);
        });
    }

    private send(client: PeerConnectionClient, type: string, data?: object) {
        const message = JSON.stringify({
            type: type,
            data: data || null,
        });

        if (client.channel.readyState != 'open') {
            client.messageQueue.push(message);
            return;
        }

        client.channel.send(message);
    }

    private createClients(lobbyId: string, users: LobbyUserObj[], currentUserId?: string): void {
        let shouldSendOffer = false;
        if (currentUserId) {
            this.userId = currentUserId;
            shouldSendOffer = true;
        }

        // filter out current user and users with existing connections 
        users
            .filter(user => user.id != this.userId && !this.clients.has(user.id))
            .forEach(m => {
                const client = this.createClient(m.id);
                client.connection.addEventListener('icecandidate', (ev) => {
                    if (ev.candidate) {
                        console.log(`[MediaSyncService] sending ice candidate for user ${client.id}`);
                        this.lobbyService.sendIceCandidate(m.id, lobbyId, ev.candidate);
                    }
                });

                if (shouldSendOffer) {
                    const offer = client.connection.createOffer();
                    offer?.then((description) => {
                        client.connection.setLocalDescription(description);

                        console.log(`[MediaSyncService] sending offer for user ${client.id} (send offer)`);
                        this.lobbyService.sendDataOffer(m.id, lobbyId, description);
                    });
                }

                this.clients.set(client.id, client);
            });
    }

    private onLobbyJoin = (res: LobbySubscriptions['join']) => {
        if (!res.data) {
            return;
        }

        const { data } = res;
        this.createClients(data.lobbyId, data.users, data.user?.id);
    };

    private onLobbyLeave = (res: LobbySubscriptions['leave']) => {
        if (!res.data) {
            return;
        }

        const { data } = res;
        const clientsToDelete: string[] = [];
        this.clients.forEach(client => {
            if (!data.users.some(q => q.id == client.id)) {
                clientsToDelete.push(client.id);
            }
        });
        clientsToDelete.forEach(clientId => {
            this.clients.delete(clientId);
        });
    };

    private onDataOffer = (res: LobbySubscriptions['data-offer']) => {
        if (!res.data) {
            return;
        }

        const { data } = res;
        const client = this.clients.get(data.from);
        if (!client) {
            return;
        }

        console.log(`[MediaSyncService] got offer for user ${client.id}`);
        client.connection.setRemoteDescription(data.description).then(() => {
            return client.connection.createAnswer();
        }).then((answer) => {
            client.connection.setLocalDescription(answer);

            console.log(`[MediaSyncService] sending answer for user ${client.id}`);
            this.lobbyService.sendDataAnswer(client.id, data.lobbyId, answer);
        }).catch(reason => console.log('Failed to create answer ' + reason));
    };

    private onDataAnswer = (res: LobbySubscriptions['data-answer']) => {
        if (!res.data) {
            return;
        }

        const { data } = res;
        const client = this.clients.get(data.from);
        if (!client) {
            return;
        }

        console.log(`[MediaSyncService] got answer for user ${client.id}`);
        client.connection.setRemoteDescription(data.description)
            .catch(reason => console.log('Failed to set answer ' + reason));
    };

    private onIceCandidate = (res: LobbySubscriptions['ice-candidate']) => {
        if (!res.data) {
            return;
        }

        const { data } = res;
        const client = this.clients.get(data.from);
        if (!client) {
            return;
        }

        console.log(`[MediaSyncService] got ice candidate for user ${client.id}`);
        client.connection.addIceCandidate(data.candidate)
            .catch(reason => console.log('Failed to add ice candidate ' + reason));
    };

    private createClient(id: string): PeerConnectionClient {
        const connection = new RTCPeerConnection({
            iceServers: [
                { urls: 'stun:stun.l.google.com:19302' },
                { urls: 'stun:stun1.l.google.com:19302' },
                { urls: 'stun:stun2.l.google.com:19302' },
                { urls: 'stun:stun3.l.google.com:19302' },
                { urls: 'stun:stun4.l.google.com:19302' }
            ]
        });
        const messageQueue: string[] = [];

        const channel = connection.createDataChannel(id, { negotiated: true, id: 0 });
        channel.addEventListener('open', () => {
            console.log(`[MediaSyncService] channel ${id} is open, ${messageQueue.length} messages in the queue`);
            messageQueue.forEach(message => {
                channel.send(message);
            });
        });
        channel.addEventListener('message', (event) => {
            const msg: { type: keyof MediaSyncSubscriptions, data: MediaSyncSubscriptions[keyof MediaSyncSubscriptions] } = JSON.parse(event.data);
            this._subscription.publish(msg.type, msg.data);
        });

        return {
            id: id,
            connection: connection,
            channel: channel,
            messageQueue: messageQueue,
        };
    }

}