export interface ISubscription<T> {
    subscribe<K extends keyof T>(type: K, callback: SubscriptionCallback<T, K>): this;
    unsubscribe<K extends keyof T>(type: K, callback: SubscriptionCallback<T, K>): this;
}

export interface ISubscriptionPublisher<T> {
    publish<K extends keyof T>(type: K, data: T[K]): void;
}

type SubscriptionCallback<T, K extends keyof T = keyof T> = (data: T[K]) => void;

export class Subscription<T> implements ISubscription<T>, ISubscriptionPublisher<T> {
    private subsciptions = new Map<keyof T, Set<SubscriptionCallback<T>>>();

    public subscribe<K extends keyof T>(type: K, callback: SubscriptionCallback<T, K>): this {
        const newCallback = callback as SubscriptionCallback<T>;
        const callbacks = this.subsciptions.get(type);

        if (!callbacks) {
            this.subsciptions.set(type, new Set([newCallback]));
            return this;
        }

        callbacks.add(newCallback);
        return this;
    }

    public unsubscribe<K extends keyof T>(type: K, callback: SubscriptionCallback<T, K>): this {
        const callbackToRemove = callback as SubscriptionCallback<T>;
        const callbacks = this.subsciptions.get(type);

        if (!callbacks) {
            return this;
        }

        callbacks.delete(callbackToRemove);
        return this;
    }

    public publish<K extends keyof T>(type: K, data: T[K]): void {
        const callbacks = this.subsciptions.get(type);
        if (!callbacks) {
            return;
        }

        callbacks.forEach(callback => callback(data as T[K]));
    }
}